{ pkgs, nix-thunk }:
self: super:
with pkgs.haskell.lib; {

  # Sometimes, I just want fast build times, and I don't care about tests,
  # documentation, or getting stuff from a cache.
  # From https://discourse.nixos.org/t/nix-haskell-development-2020/6170/2
  mkDerivation = args: super.mkDerivation (args // {
      doCheck = false;
      doHaddock = false;
      enableLibraryProfiling = false;
      enableExecutableProfiling = false;
      # jailbreak = true; # Maybe a bit too much hehe
      # Disable compiler optimizations for faster compilation:
      # buildFlags = [ "--ghc-options=-O0" ];
    });

  # Use our fork of android-activity which adds support for VIEW intents.
  android-activity = import
    (nix-thunk.thunkSource ./dep/android-activity)
    { inherit (pkgs.buildPackages) jdk;
      inherit (self) callPackage callCabal2nix;
    };

  # AND OUR FORK OF REFLEX-DOM.
  reflex-dom = super.reflex-dom.overrideAttrs (old: {
    src = "${nix-thunk.thunkSource ./dep/reflex-dom}/reflex-dom";
  });

  # Here, we get the tahoe-chk and tahoe-ssk packages from their thunk and use
  # callCabal2nix to get nix derivations to build them.
  tahoe-chk = self.callCabal2nix
    "tahoe-chk"
    (nix-thunk.thunkSource ./dep/tahoe-chk)
    {};

  # magic-wormhole depends on this extra crypto library.
  saltine = super.saltine.overrideAttrs (old : {
    nativeBuildInputs = old.nativeBuildInputs ++ [ pkgs.pkg-config ];
  });

  tahoe-ssk = self.callCabal2nix
    "tahoe-ssk"
    (nix-thunk.thunkSource ./dep/tahoe-ssk)
    {};

  tahoe-capabilities = self.callCabal2nix
    "tahoe-capability"
    (nix-thunk.thunkSource ./dep/tahoe-capabilities)
    {};

  tahoe-directory = self.callCabal2nix
    "tahoe-directory"
    (nix-thunk.thunkSource ./dep/tahoe-directory)
    {};

  # Similarly, get a package implementing Great Black Swamp.
  tahoe-great-black-swamp = self.callCabal2nix
    "tahoe-great-black-swamp"
    (nix-thunk.thunkSource ./dep/tahoe-great-black-swamp)
    {};

  # this is marked as broken, because there's no python to run the
  # integration tests; so we override to git our own pattern that
  # skips (just) the integration test
  spake2 = (self.callCabal2nix
    "spake2"
    (nix-thunk.thunkSource ./dep/haskell-spake2)
    {}).overrideDerivation (
      oldAttrs: {
        broken=false;
        dontCheck=true;
        checkPhase=''
set -x
runHook preCheck
./Setup test --test-options '--pattern "$2 != \"Integration\""'
runHook postCheck
'';
      }
    );

  # magic-wormhole base library (similar to above, we skip the
  # integration tests because no pyhthon)

  magic-wormhole = (self.callCabal2nix
    "haskell-magic-wormhole"
    (nix-thunk.thunkSource ./dep/haskell-magic-wormhole)
    {}).overrideDerivation (
      oldAttrs: {
        broken=false;
        dontCheck=true;
        checkPhase=''
set -x
runHook preCheck
./Setup test --test-options '--pattern "$2 != \"Integration\""'
runHook postCheck
'';
      }
    );

  # And the integration library that pulls the pieces together and presents a
  # high-level interface.
  gbs-downloader = self.callCabal2nix
    "gbs-downloader"
    (nix-thunk.thunkSource ./dep/gbs-downloader)
    {};

  magic-foldr = self.callCabal2nix
    "magic-foldr"
    (nix-thunk.thunkSource ./dep/magic-foldr)
    {};

  # We also ended up needing an override of the base32 library, which we
  # obtain from Hackage.
  base32 = self.callHackageDirect {
    pkg = "base32";
    ver = "0.2.1.0";
    sha256 = "04glpmwp50qi29h8cb5j0w1rz0ww30nw4xgj2a3l7zh1gprhwj89";
  } {};

  # The test suite uses >35GB of RAM.
  servant-docs = dontCheck super.servant-docs;

  # A dependency of servant that's missing from reflex-platform.
  hspec-wai = self.callHackageDirect {
    pkg = "hspec-wai";
    ver = "0.9.2";
    sha256 = "sha256-9un9jdqLz/wrLxL5i6GGeVaievlQCWLi8Pr1PgNLJ1Q=";
  } {};

  # nixpkgs doesn't capture the Android dependencies for this package due to
  # conditionals in the cabal file or maybe CPP usage in the source.  Try to
  # fix it up.
  #
  # Possibly we should split tahoe-great-black-swamp into separate client and
  # server libraries so we don't need this dependency on Android at all.  This
  # is easier for now.
  simple-sendfile = self.callPackage ./simple-sendfile.nix { };
}
